/////////////////////////////////////////////////////////////////////
// class to show shell contextmenu of files/folders/shell objects
// developed by R. Engels 2003
//
// Modified by Erik Christiansson to fit in IconDesk
/////////////////////////////////////////////////////////////////////

#include "ShellContextMenu.h"

#define MIN_ID 1
#define MAX_ID 10000

IContextMenu2 * g_IContext2 = NULL;
IContextMenu3 * g_IContext3 = NULL;

ShellContextMenu::ShellContextMenu()
{
	m_psfFolder = NULL;
	m_pidlArray = NULL;
	_hMenu = NULL;
}

ShellContextMenu::~ShellContextMenu()
{
	// free all allocated datas
	if(m_psfFolder && bDelete) {
		m_psfFolder->Release();
	}
	m_psfFolder = NULL;
	PIDLFreeArray(m_pidlArray);
	m_pidlArray = NULL;

	if(_hMenu) {
		DestroyMenu(_hMenu);
	}
}

// this functions determines which version of IContextMenu is avaibale for those objects (always the highest one)
// and returns that interface
BOOL ShellContextMenu::GetContextMenu (void ** ppContextMenu, int & iMenuType)
{
	*ppContextMenu = NULL;
	LPCONTEXTMENU icm1 = NULL;
	
	// first we retrieve the normal IContextMenu interface (every object should have it)
	m_psfFolder->GetUIObjectOf(NULL, nItems, (LPCITEMIDLIST *) m_pidlArray, IID_IContextMenu, NULL, (void**) &icm1);

	if(icm1) {	// since we got an IContextMenu interface we can now obtain the higher version interfaces via that
		if(icm1->QueryInterface (IID_IContextMenu3, ppContextMenu) == NOERROR) {
			iMenuType = 3;
		}
		else if(icm1->QueryInterface (IID_IContextMenu2, ppContextMenu) == NOERROR) {
			iMenuType = 2;
		}

		if(*ppContextMenu) {
			icm1->Release(); // we can now release version 1 interface, cause we got a higher one
		}
		else {	
			iMenuType = 1;
			*ppContextMenu = icm1;	// since no higher versions were found
		}							// redirect ppContextMenu to version 1 interface
	}
	else {
		return FALSE;	// something went wrong
	}
	
	return TRUE; // success
}

LRESULT CALLBACK ShellContextMenu::HookWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{ 
	case WM_MENUCHAR:	// only supported by IContextMenu3
		if(g_IContext3) {
			LRESULT lResult = 0;
			g_IContext3->HandleMenuMsg2 (message, wParam, lParam, &lResult);
			return lResult;
		}
		break;

	case WM_DRAWITEM:
	case WM_MEASUREITEM:
		if(wParam) { 
			break; // if wParam != 0 then the message is not menu-related
		}
  
	case WM_INITMENUPOPUP:
		if(g_IContext2) {
			g_IContext2->HandleMenuMsg (message, wParam, lParam);
		}
		else {	// version 3
			g_IContext3->HandleMenuMsg (message, wParam, lParam);
		}
		return (message == WM_INITMENUPOPUP ? 0 : TRUE); // inform caller that we handled WM_INITPOPUPMENU by ourself
		break;

	default:
		break;
	}

	// call original WndProc of window to prevent undefined bevhaviour of window
	return ::CallWindowProc((WNDPROC)GetProp(hWnd, TEXT("OldWndProc")), hWnd, message, wParam, lParam);
}

UINT ShellContextMenu::ShowContextMenu(LPITEMIDLIST pidl, HWND hWnd)
{
	SetObjects(pidl);

	POINT pt;
	GetCursorPos(&pt);

	int iMenuType = 0;	// to know which version of IContextMenu is supported
	LPCONTEXTMENU pContextMenu;	// common pointer to IContextMenu and higher version interface
   
	if(!GetContextMenu((void**)&pContextMenu, iMenuType)) {
		return 0;	// something went wrong
	}

	if(!_hMenu) {
		DestroyMenu(_hMenu);
		_hMenu = CreatePopupMenu();
	}

	// lets fill the our popupmenu  
	pContextMenu->QueryContextMenu(_hMenu, 0, MIN_ID, MAX_ID, CMF_NORMAL | CMF_EXPLORE);
 
	// subclass window to handle menurelated messages in ShellContextMenu 
	WNDPROC OldWndProc;
	if(iMenuType > 1) {	// only subclass if its version 2 or 3
		OldWndProc = (WNDPROC)SetWindowLong(hWnd, GWL_WNDPROC, (DWORD)HookWndProc);
		if(iMenuType == 2) {
			g_IContext2 = (LPCONTEXTMENU2)pContextMenu;
		}
		else {	// version 3
			g_IContext3 = (LPCONTEXTMENU3)pContextMenu;
		}
	}
	else {
		OldWndProc = NULL;
	}

	UINT idCommand = TrackPopupMenu(_hMenu, TPM_RETURNCMD | TPM_LEFTALIGN, pt.x, pt.y, NULL, hWnd, NULL);

	if(OldWndProc) { // unsubclass
		SetWindowLong(hWnd, GWL_WNDPROC, (DWORD)OldWndProc);
	}

	if(idCommand >= MIN_ID && idCommand <= MAX_ID) {	// see if returned idCommand belongs to shell menu entries
		InvokeCommand(pContextMenu, idCommand - MIN_ID);	// execute related command
		idCommand = 0;
	}
	
	pContextMenu->Release();
	g_IContext2 = NULL;
	g_IContext3 = NULL;

	return idCommand;
}

void ShellContextMenu::InvokeCommand(LPCONTEXTMENU pContextMenu, UINT idCommand)
{
	CMINVOKECOMMANDINFO cmi = {0};
	cmi.cbSize = sizeof(CMINVOKECOMMANDINFO);
	cmi.lpVerb = (LPSTR)MAKEINTRESOURCE(idCommand);
	cmi.nShow = SW_SHOWNORMAL;
	
	pContextMenu->InvokeCommand(&cmi);
}

// only one full qualified PIDL has been passed
void ShellContextMenu::SetObjects(LPITEMIDLIST pidl)
{
	// free all allocated datas
	if(m_psfFolder && bDelete) {
		m_psfFolder->Release();
	}
	m_psfFolder = NULL;
	PIDLFreeArray(m_pidlArray);
	m_pidlArray = NULL;

	// full qualified PIDL is passed so we need
	// its parent IShellFolder interface and its relative PIDL to that
	LPITEMIDLIST pidlItem = NULL;
	SHBindToParent((LPCITEMIDLIST)pidl, IID_IShellFolder, (void **)&m_psfFolder, (LPCITEMIDLIST*)&pidlItem);

	m_pidlArray = (LPITEMIDLIST *)malloc(sizeof(LPITEMIDLIST));	// allocate ony for one elemnt
	m_pidlArray[0] = PIDLCopy(pidlItem);


	// now free pidlItem via IMalloc interface (but not m_psfFolder, that we need later
	LPMALLOC lpMalloc = NULL;
	SHGetMalloc(&lpMalloc);
	lpMalloc->Free(pidlItem);
	lpMalloc->Release();

	nItems = 1;
	bDelete = TRUE;	// indicates that m_psfFolder should be deleted by ShellContextMenu
}

HMENU ShellContextMenu::GetMenu()
{
	if(!_hMenu) {
		_hMenu = CreatePopupMenu();	// create the popupmenu (its empty)
	}
	return _hMenu;
}

// this is workaround function for the Shell API Function SHBindToParent
// SHBindToParent is not available under Win95/98
HRESULT ShellContextMenu::SHBindToParentEx(LPCITEMIDLIST pidl, REFIID riid, VOID **ppv, LPCITEMIDLIST *ppidlLast)
{
	HRESULT hr = 0;
	if(!pidl || !ppv) {
		return E_POINTER;
	}
	
	int nCount = PIDLGetItemCount(pidl);
	if(nCount == 0)	{ // desktop pidl of invalid pidl
		return E_POINTER;
	}

	IShellFolder *psfDesktop = NULL;
	SHGetDesktopFolder(&psfDesktop);
	if(nCount == 1) {	// desktop pidl
		if ((hr = psfDesktop->QueryInterface(riid, ppv)) == S_OK) {
			if(ppidlLast) {
				*ppidlLast = PIDLCopy(pidl);
			}
		}
		psfDesktop->Release();
		return hr;
	}

	LPBYTE pRel = PIDLGetPos(pidl, nCount - 1);
	LPITEMIDLIST pidlParent = NULL;
	pidlParent = PIDLCopy(pidl, pRel - (LPBYTE)pidl);
	IShellFolder * psfFolder = NULL;
	
	if((hr = psfDesktop->BindToObject(pidlParent, NULL, __uuidof(psfFolder), (void **)&psfFolder)) != S_OK) {
		free(pidlParent);
		psfDesktop->Release ();
		return hr;
	}
	if((hr = psfFolder->QueryInterface(riid, ppv)) == S_OK) {
		if(ppidlLast) {
			*ppidlLast = PIDLCopy((LPCITEMIDLIST) pRel);
		}
	}
	free(pidlParent);
	psfFolder->Release();
	psfDesktop->Release();
	return hr;
}
