/*
This is a part of the IconDesk source code.

Copyright (C) 2003 Erik Christiansson, aka Sci
www.alphafish.com
erik@alphafish.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "pidl.h"

LPITEMIDLIST PIDLAppend(LPCITEMIDLIST pidlBase, LPCITEMIDLIST pidlAdd, LPMALLOC pMalloc)
{
	if(pidlBase == NULL) {
        return PIDLCopy(pidlAdd);
	}
	if(pidlAdd == NULL) {
        return PIDLCopy(pidlBase);
	}
    
    LPITEMIDLIST pidlNew;

    UINT cb1 = PIDLGetSize(pidlBase) - sizeof(pidlBase->mkid.cb);
    UINT cb2 = PIDLGetSize(pidlAdd);

    pidlNew = (LPITEMIDLIST)pMalloc->Alloc(cb1 + cb2);
    if(pidlNew) {
        CopyMemory(pidlNew, pidlBase, cb1);
        CopyMemory(((LPSTR)pidlNew) + cb1, pidlAdd, cb2);
    }
    return pidlNew;
}

LPITEMIDLIST GetNextItemID(LPCITEMIDLIST pidl) 
{ 
	// Check for valid pidl.
	if(pidl == NULL)
		return NULL;

	// Get the size of the specified item identifier. 
	int cb = pidl->mkid.cb; 

	// If the size is zero, it is the end of the list. 
	if (cb == 0) 
		return NULL; 

	// Add cb to pidl (casting to increment by bytes). 
	pidl = (LPITEMIDLIST) (((LPBYTE) pidl) + cb); 

	// Return NULL if it is null-terminating, or a pidl otherwise. 
	return (pidl->mkid.cb == 0) ? NULL : (LPITEMIDLIST) pidl; 
} 

/**
 * PIDLGetSize
 * Computes the total size of an item ID list.
 */
UINT PIDLGetSize(LPCITEMIDLIST pidl)
{
    UINT cbTotal = 0;
    if (pidl)
    {
        cbTotal += sizeof(pidl->mkid.cb);    // Null terminator
        while (pidl)
        {
            cbTotal += pidl->mkid.cb;
            pidl = GetNextItemID(pidl);
        }
    }
    return cbTotal;
}

LPITEMIDLIST PIDLCopy(LPCITEMIDLIST pidl, UINT cb)
{
    LPMALLOC pMalloc;
    HRESULT hr;
    hr = SHGetMalloc(&pMalloc);

	LPITEMIDLIST pidlTemp = (LPITEMIDLIST) pidl;
	
	// Calculate size of list.
	if(cb == -1) {
	    cb = PIDLGetSize(pidl);
	}

    LPITEMIDLIST pidlRet = (LPITEMIDLIST)pMalloc->Alloc(cb);
	if(pidlRet) {
        CopyMemory(pidlRet, pidl, cb);
	}

    pMalloc->Release();

    return pidlRet;
}

/**
 * PIDLGetItemCount
 * Computes the number of item IDs in an item ID list.
 */
UINT PIDLGetItemCount(LPCITEMIDLIST pidl)
{
	USHORT nLen;
	UINT nCount;
	nCount = 0;
	while((nLen = pidl->mkid.cb) != 0) {
		pidl = GetNextItem (pidl);
		nCount++;
	}
	return nCount;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//  FUNCTION:       GetNextItem
//
//  DESCRIPTION:    Finds the next item in an item ID list.
//
//  INPUT:          pidl = Pointer to an item ID list.
//
//  RETURNS:        Pointer to the next item.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

LPITEMIDLIST GetNextItem (LPCITEMIDLIST pidl)
{
	USHORT nLen;

	if ((nLen = pidl->mkid.cb) == 0)
		return NULL;

	return (LPITEMIDLIST) (((LPBYTE) pidl) + nLen);
}

LPBYTE PIDLGetPos(LPCITEMIDLIST pidl, int nPos)
{
	if(!pidl) {
		return 0;
	}
	int nCount = 0;
	
	BYTE * pCur = (BYTE *)pidl;
	while(((LPCITEMIDLIST) pCur)->mkid.cb) {
		if(nCount == nPos) {
			return pCur;
		}
		nCount++;
		pCur += ((LPCITEMIDLIST)pCur)->mkid.cb;	// + sizeof(pidl->mkid.cb);
	}
	if(nCount == nPos) {
		return pCur;
	}
	return NULL;
}

void PIDLFreeArray(LPITEMIDLIST *pidlArray)
{
	if(!pidlArray) {
		return;
	}

	int iSize = _msize(pidlArray) / sizeof(LPITEMIDLIST);

	for(int i = 0; i < iSize; i++) {
		free(pidlArray[i]);
	}
	free(pidlArray);
}
